Slammer = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "AceDebug-2.0", "AceHook-2.1", "CandyBar-2.0", "Sink-1.0")
-- embedded libs
local L = AceLibrary("AceLocale-2.2"):new("Slammer")
local BS = AceLibrary("Babble-Spell-2.2")

Slammer.dataSetup = {}

local colors = {"white", "black", "blue", "magenta", "cyan", "green", "yellow", "orange", "red", "gray", "forest", "maroon", "navy", "olive", "purple", "teal"}

local options = {
    type = "group",
    args = {
        anchor = {
            name = "Anchor", desc = "Shows the dragable anchor.", type = "execute", 
            func = "ToggleAnchor"
        },
        update = {
            name = "Update anchor", desc = "Update the dragable size.", type = "execute", 
            func = "UpdateAnchorSize"
        },
        height = {
            name = "Bar Height", desc="Set the bar height.", type = "range",
            get = function () return Slammer.db.profile.barheight end,
            set = function (v) Slammer.db.profile.barheight = v; Slammer:UpdateAnchorSize() end,
            min = 1, max = 64,
        },
        width = {
            name = "Bar Width", desc="Set the bar width.", type = "range",
            get = function () return Slammer.db.profile.barwidth end,
            set = function (v) Slammer.db.profile.barwidth = v; Slammer:UpdateAnchorSize() end,
            min = 50, max = 1000,
        },
        bloodfrenzy = {
            name = "Blood Frenzy", desc = "Enable or disable Blood Frenzy damage tracking.", type = "toggle", 
            get = function () return Slammer.db.profile.showbloodfrenzy end,
            set = function (v) Slammer.db.profile.showbloodfrenzy = v end,
            map = { [false] = "Disabled", [true] = "Enabled" }
        },
        interrupt = {
            name = "Interrupt", desc = "Enable or disable global interrupt tracking.", type = "toggle", 
            get = function () return Slammer.db.profile.showinterrupt end,
            set = function (v) Slammer.db.profile.showinterrupt = v end,
            map = { [false] = "Disabled", [true] = "Enabled" }
        },
        nag = {
            name = "Nag", desc = "Enable or disable nags.", type = "toggle", 
            get = function () return Slammer.db.profile.shownag end,
            set = function (v) Slammer.db.profile.shownag = v end,
            map = { [false] = "Disabled", [true] = "Enabled" }
        },
        debuffCheck = {
            name = "Debuff Check", desc = "Enable or disable debuff checks.", type = "toggle", 
            get = function () return Slammer.db.profile.showdebuffcheck end,
            set = function (v) Slammer.db.profile.showdebuffcheck = v end,
            map = { [false] = "Disabled", [true] = "Enabled" }
        },
        reversebar = {
            name = "Reverse bar", desc = "Reverse bar direction.", type = "toggle", 
            get = function () return Slammer.db.profile.reversebar end,
            set = function (v) Slammer.db.profile.reversebar = v end,
            map = { [false] = "Right to left", [true] = "Left to right." }
        },
        tclaptime = {
            name = "Thunderclap nag time.", desc="Use -1 to disable this nag.", type = "range",
            get = function () return Slammer.db.profile.tclaptime end,
            set = function (v) Slammer.db.profile.tclaptime = v end,
            min = -1, max = 30,
        },
        sundertime = {
            name = "Sunder Armor nag time.", desc="Use -1 to disable this nag.", type = "range",
            get = function () return Slammer.db.profile.sundertime end,
            set = function (v) Slammer.db.profile.sundertime = v end,
            min = -1, max = 30,
        },
        demotime = {
            name = "Demoralizing shout nag time.", desc="Use -1 to disable this nag.", type = "range",
            get = function () return Slammer.db.profile.demotime end,
            set = function (v) Slammer.db.profile.demotime = v end,
            min = -1, max = 45,
        },
        bleedtime = {
            name = "Bleed effect nag time.", desc="Use -1 to disable this nag.", type = "range",
            get = function () return Slammer.db.profile.bleedtime end,
            set = function (v) Slammer.db.profile.bleedtime = v end,
            min = -1, max = 21,
        },
        buffshouttime = {
            name = "Buffing shout nag time.", desc="Use -1 to disable this nag.", type = "range",
            get = function () return Slammer.db.profile.buffshouttime end,
            set = function (v) Slammer.db.profile.buffshouttime = v end,
            min = -1, max = 180,
        },
    },
}

local lastOffset = 0
local offsetAge = 0
local lastSwingResetTime = 0
local expectedNextHitTime = GetTime()
local extraAttackTimeStamp = GetTime()
local lastSelfShoutTime = 0
local lastSelfShout = nil -- can be nil, "BS" or "CS"
local BFDamage = 0
local BFDamageDowntime = 0
local BFActiveGUID = {}
local BFDamageSource = {}
local startTime = 0
local wantedList = {}

local wcf = CreateFrame("Button", nil, UIParent) -- wcf: wasted cycle frame

function Slammer:OnInitialize()
    local paint = AceLibrary("PaintChips-2.0")
    paint:RegisterColor("gray", 0.5,0.5,0.5)
    paint:RegisterColor("forest", 0.0,0.5,0.0)
    paint:RegisterColor("maroon", 0.5,0.0,0.0)
    paint:RegisterColor("navy", 0.0,0.0,0.5)
    paint:RegisterColor("olive", 0.5,0.5,0.0)
    paint:RegisterColor("purple", 0.5,0.0,0.5)
    paint:RegisterColor("teal", 0.0,0.5,0.5)

    local hr = 3 --headroom
    local sc = 1 --stance change
    local gcd = 1.5 -- global cooldown
    local defaults = {
            barposition = {},
            barwidth = 690,
            barheight = 24,
            shownag = true,
            showinterrupt = false,
            showbloodfrenzy = true,
            showdebuffcheck = true,
            reversebar = false,
            tclaptime =  4   + hr + sc, --one cooldown time + 2 sec of headroom + 1 sec of stance change
            sundertime = gcd + hr,
            demotime =   gcd + hr,
            bleedtime =  gcd + hr + sc,
            buffshouttime = hr,
    }
    Slammer:RegisterDB("SlammerDB")
    Slammer:RegisterDefaults('profile', defaults)
    Slammer:RegisterChatCommand({'/slam', '/slammer'}, options)

    -- create the anchor frame
    self.anchor = self:CreateAnchor("Slammer", 0,1,0)

    wcf:SetWidth(300)
    wcf:SetHeight(self.db.profile.barheight * 2)
    wcf.owner = self
    wcf:ClearAllPoints()
    wcf:SetPoint("BOTTOM", self.anchor, "TOP",0,0)
    wcf.Text = wcf:CreateFontString(nil, "OVERLAY")
    wcf.Text:SetFontObject(GameFontNormalSmall)
    wcf.Text:SetTextColor(1, 1, 1, 1)
    wcf.Text:SetWidth(200)
    wcf.Text:SetHeight(self.db.profile.barheight)
    wcf.Text:SetJustifyH("CENTER")
    wcf.Text:SetJustifyV("MIDDLE")
    wcf.Text:ClearAllPoints()
    wcf.Text:SetPoint("BOTTOM", wcf, "BOTTOM")
    wcf:EnableMouse(false)
    
    hitTimeStamp = GetTime()

    self:RegisterCandyBarGroup("Slammer")
    self:SetCandyBarGroupPoint("Slammer", "BOTTOMLEFT", self.anchor, "BOTTOMLEFT", -self.db.profile.barheight, 0)    -- 16 = height
end

function Slammer:OnEnable()
    self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");

    self:RegisterEvent("COMBAT_TEXT_UPDATE");
    self:RegisterEvent("PLAYER_REGEN_ENABLED") -- leaving combat
    self:RegisterEvent("PLAYER_REGEN_DISABLED") -- entering combat
    self:RegisterEvent("PLAYER_TARGET_CHANGED")
    self:RegisterEvent("ITEM_LOCK_CHANGED")

    --self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED",  function (arg1, arg2, arg3) self:SPELLCAST_SUCEEDED(arg1, arg2, arg3) end);
end

function Slammer:OnDisable()
    self:UnregisterAllEvents()
    self:UnhookAll()
end

--<< ====================================================================== >>--
-- Anchor Processing                                                          --
--<< ====================================================================== >>--

-- Toggles the vis of the anchors
function Slammer:ToggleAnchor()
    if self.anchor:IsVisible() then
        self.anchor:Hide()
    else
        self.anchor:Show()
    end
end

function Slammer:UpdateAnchorSize()
    self.anchor:SetWidth(self.db.profile.barwidth)
    self.anchor:SetHeight(self.db.profile.barheight)
end

-- Creates the anchor frames
function Slammer:CreateAnchor(text, cRed, cGreen, cBlue)
    local f = CreateFrame("Button", nil, UIParent)
    f:SetWidth(self.db.profile.barwidth)
    f:SetHeight(self.db.profile.barheight)

    f.owner = self

    if self.db.profile.barposition.x and self.db.profile.barposition.y and self.db.profile.barposition.anchor then
        f:ClearAllPoints()
        f:SetPoint(self.db.profile.barposition.anchor, UIParent, self.db.profile.barposition.anchor, self.db.profile.barposition.x, 
            self.db.profile.barposition.y)
    else
        f:SetPoint("CENTER", UIParent, "CENTER", 0, 50)
    end

    f:SetScript("OnDragStart", function() this:StartMoving() end )
    f:SetScript("OnDragStop",
        function()
            this:StopMovingOrSizing()
            local p, t, v, x, y = this:GetPoint()
            this.owner.db.profile.barposition.x = math.floor(x)
            this.owner.db.profile.barposition.y = math.floor(y)
            this.owner.db.profile.barposition.anchor = v
        end)

--  f:SetBackdrop({bgFile = "Interface\DialogFrame\UI-DialogBox-Background", tile = true, tileSize = 32, insets = { left = 0, right = 0, top = 0, bottom = 0 }})
--                                            edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
--                                            tile = false, tileSize = 16, edgeSize = 16,
--                                            insets = { left = 5, right =5, top = 5, bottom = 5 }})
    f:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background",
                                            edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
                                            tile = false, tileSize = 16, edgeSize = 16,
                                            insets = { left = 0, right = 0, top = 0, bottom = 0 }})

    f:SetBackdropColor(cRed,cGreen,cBlue,.6)
    f:SetMovable(true)
    f:RegisterForDrag("LeftButton")

    f.Text = f:CreateFontString(nil, "OVERLAY")
    f.Text:SetFontObject(GameFontNormalSmall)
    f.Text:ClearAllPoints()
    f.Text:SetTextColor(1, 1, 1, 1)
    f.Text:SetWidth(200)
    f.Text:SetHeight(self.db.profile.barheight)
    f.Text:SetPoint("TOPLEFT", f, "TOPLEFT")
    f.Text:SetJustifyH("CENTER")
    f.Text:SetJustifyV("MIDDLE")
    f.Text:SetText(text)

    f:Hide()

    return f
end

local spiritShockInterruptID = 1
--<< ====================================================================== >>--
-- Event/Hook processing                                                         --
--<< ====================================================================== >>--
function Slammer:COMBAT_LOG_EVENT_UNFILTERED(...)
    local timestamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName = select(1, ...);
    
    local player_GUID = UnitGUID("player");
    local target_GUID = UnitGUID("target");
    local channel = UnitHealth("PLAYER") > 0 and "SAY" or "RAID_WARNING"
    -- Global interrupt message
    if self.db.profile.showinterrupt and eventType == "SPELL_INTERRUPT" then
        extraSpellID, extraSpellName, extraSpellSchool = select(1+3+8,...)
        self:Print(srcName .. " interrupted " .. dstName.. "'s " ..extraSpellName.." with "..spellName..", locking "..extraSpellSchool..".");
        -- SendChatMessage(srcName .. " interrupted " .. dstName.. "'s " ..extraSpellName.." with "..spellName..", locking "..extraSpellSchool..".", channel);
    end
    
    if eventType == "SPELL_CAST_START" and spellName == "Spirit Shock" then 
        SendChatMessage("Interrupt "..spiritShockInterruptID, channel);
        spiritShockInterruptID = spiritShockInterruptID == 3 and 1 or spiritShockInterruptID + 1
    end
    
    if eventType == "SPELL_CAST_SUCCESS" and spellName == "Flame Blast" and srcGUID == target_GUID then
        self:FlameBlastHappened();
    end

    -- Swing timer management
    if srcGUID == player_GUID then
        if eventType:find("SWING") or
        (eventType == "SPELL_DAMAGE" and (spellName == "Heroic Strike" or spellName == "Cleave")) then
            self:NormalHit(nil);
        elseif (eventType == "SPELL_DAMAGE" or eventType == "SPELL_MISSED") and (spellName == "Slam") then
            self:NormalHit(true);
        elseif (eventType == "SPELL_DAMAGE" or eventType == "SPELL_MISSED") and (spellName == "Mortal Strike" or spellName == "Bloodthirst" or spellName == "Shield Slam")  then
            self:SalvTankingCheck();
        end
    -- Shout tracking
        if eventType == "SPELL_CAST_SUCCESS" then 
            if spellName == "Battle Shout" then
                lastSelfShout = "BS"
                lastSelfShoutTime = GetTime()
            elseif spellName == "Commanding Shout" then
                lastSelfShout = "CS"            
                lastSelfShoutTime = GetTime()
            end
        end
        -- Hit-slam delay update
        if eventType == "SPELL_CAST_START" and spellName == "Slam" then
            self:HitSlamDelay()
        end
        -- DrDamage mitigation update
        if DrDamage and eventType == "SPELL_DAMAGE" and dstGUID == target_GUID and 
        (
        spellName == "Thunder Clap" or 
        spellName == "Hamstring" or 
        spellName == "Bloodthirst" or 
        spellName == "Victory Rush" or 
        spellName == "Pummel" or
        spellName == "Mocking Blow" or
        spellName == "Shield Bash" or
        spellName == "Intercept Stun"
        ) then
            -- local amount, school, resisted, blocked, absorbed, critical, glancing, crushing = select(1+3+8,...)
            local amount, _, _, _, _, critical, _, _ = select(1+3+8,...)
            self:UpdateDrDamageMitigation(spellName, amount, critical, UnitLevel("TARGET"))
        end
    end
    -- Next swing time change
    if dstGUID == player_GUID then
        -- If missType parried
        local missType = nil
        if eventType == "SWING_MISSED" then
            missType = select(1+8,...)
        elseif eventType == "SPELL_MISSED" then
            missType = select(1+3+8,...)
        end
        if missType and missType == "PARRY" then
            self:Parried();
        end
    end
    
    -- Overpower alert
    if srcGUID == player_GUID then
        local missType = nil
        if eventType == "SWING_MISSED" then
            missType = select(1+8,...)
        elseif eventType == "SPELL_MISSED" then
            missType = select(1+3+8,...)
        end
        if missType and missType == "DODGE" then
            local _, playerClass = UnitClass("player")
            local _, playerRace  = UnitRace("player")
            if playerClass == "WARRIOR" and playerRace == "Gnome" and UnitSex("player") == 3 then
                PlaySoundFile("Sound\\Character\\Gnome\\GnomeVocalFemale\\GnomeFemaleCharge01.wav");
            end
        end
    end
    
    if self.db.profile.showbloodfrenzy then
        -- Bleed tracking for blood frenzy
        if eventType == "SPELL_AURA_APPLIED" and spellName == "Deep Wound" then -- another way to check bleedability
            if not BFActiveGUID[dstGUID] then
                BFActiveGUID[dstGUID] = 0 -- We are unsure if it is our bleed effect
            end        
        end
        if srcGUID == player_GUID and BFActiveGUID[dstGUID] and BFActiveGUID[dstGUID] == 0 and eventType == "SPELL_PERIODIC_DAMAGE" and spellName == "Deep Wound" then        
            --self:Print("Confirming our Deep Wound for the first time, adding 9 sec of bleedage on "..dstName.." because of Deep Wound.");
            local moreTime = 9
            --self:Print("Adding target "..player_GUID.." more time ("..moreTime..").")
            BFActiveGUID[dstGUID] = math.max(BFActiveGUID[dstGUID], GetTime() + moreTime)
        end
        if srcGUID == player_GUID and eventType == "SPELL_CAST_SUCCESS" then
            local moreTime = nil
            if spellName == "Rend" then
                moreTime = 21
            elseif spellName == "Deep Wound" then -- never used, deep wound doesn't fire a SPELL_CAST_SUCCESS event
                moreTime = 12
            end
            if moreTime then
                --self:Print("Adding target "..dstName.." more time ("..moreTime..").")
                if BFActiveGUID[dstGUID] then 
                    BFActiveGUID[dstGUID] = math.max(BFActiveGUID[dstGUID], GetTime() + moreTime)
                else
                    BFActiveGUID[dstGUID] = GetTime() + moreTime
                end
            end
        end
        if srcGUID == player_GUID and BFActiveGUID[dstGUID] then -- Has bleeded at some point, so if we crit and have deep wound, it's a 12 seconds DoT
            local critical = nil;
            if eventType == "SWING_DAMAGE" then
                critical = select(1+8+5,...)
            elseif eventType == "SPELL_DAMAGE" and spellName ~= "Thunder Clap" then --Should check if uncommon spells like intercept, pummel, shield bash, etc. causes deep wound
                critical = select(1+3+8+5,...)
            end
            if critical and self:HasDeepWounds() then
                --self:Print("Adding 12 sec of bleedage on "..dstName.." because of Deep Wound.");
                local moreTime = 12
                --self:Print("Adding target "..player_GUID.." more time ("..moreTime..").")
                BFActiveGUID[dstGUID] = math.max(BFActiveGUID[dstGUID], GetTime() + moreTime)
            end
        end
    
        local amount, school;
        if eventType == "SWING_DAMAGE" then
            amount, school = select(1+8,...)
        elseif eventType == "SPELL_DAMAGE" or eventType == "RANGE_DAMAGE" or eventType == "SPELL_PERIODIC_DAMAGE" then
            amount, school = select(1+3+8,...)
        end
        if school == 1 then
            self:AddBFDamage(srcName,amount, dstGUID)
        end
    end
    -- for screenshots
    if eventType == "UNIT_DIED" then 
        if wantedList[dstGUID] then
            self:PreBossDied(dstName);
--            HideUIPanel(FriendsFrame);
--            HideUIPanel(CharacterFrame);
--            PanelTemplates_SetTab(FriendsFrame, 5);
--            PanelTemplates_SetTab(CharacterFrame, 1);
--            self:ScheduleEvent(function () self:BossDied(dstName) end, 0.5);
        end
    end
end

function Slammer:PLAYER_REGEN_ENABLED()
    local timePeriod = GetTime() - startTime
    if BFDamage ~= 0 and BFDamageDowntime ~= 0 then
        local totalBFDamage = BFDamage+BFDamageDowntime
        self:Print( string.format("Blood frenzy damage: %0.0f/%0.0f (%0.1f%%) over %0.1f sec (%0.1f/%0.1f)",BFDamage, totalBFDamage, 100*BFDamage/totalBFDamage, timePeriod, BFDamage/timePeriod, totalBFDamage/timePeriod) )
    else
        if BFDamage ~= 0 then
            self:Print( string.format("Blood frenzy damage: %0.0f over %0.1f sec (%0.1f)",BFDamage, timePeriod, BFDamage/timePeriod) )
        end
        if BFDamageDowntime ~= 0 then
            self:Print(string.format("Blood frenzy damage lost to downtime: %0.0f over %0.1f sec (%0.1f)", BFDamageDowntime, timePeriod, BFDamageDowntime/timePeriod) )
        end
    end
    BFDamage = 0
    BFDamageDowntime = 0
    playerDamage = BFDamageSource[UnitName("PLAYER")]
    
    -- Idividual character display
    local sortedList = {} 
    local numberDisplayed = 0
    local nameList = ""
    local listTotal = 0
    local listSize = 0
    for key,val in pairs(BFDamageSource) do -- Fills a temporary table with for sorting. Key is an index, value is a pair with [1] being the name and [2] being the value
        if val ~= 0 then 
            table.insert(sortedList, {key,val}); 
            BFDamageSource[key] = 0 
        end
    end
    table.sort(sortedList, function(a, b) return a[2] > b[2] end)   -- Highest value first
    for _, curPair in ipairs(sortedList) do
        numberDisplayed = numberDisplayed + 1
        if numberDisplayed < 3 then
            if (curPair[1] ~= UnitName("PLAYER")) then
                self:Print( string.format("%s got %0.0f extra damage from Blood Frenzy (%0.1f)", curPair[1], curPair[2], curPair[2]/timePeriod) )
            else
                numberDisplayed = numberDisplayed - 1
            end
        else
            listSize = listSize +1            
            listTotal = listTotal + curPair[2]
            nameList = nameList .. ", " .. curPair[1]
        end
    end
    if listSize > 1 then
        self:Print( string.format("And %d others: %s. Total: %0.0f (%0.1f) Average: %0.1f (%0.1f)", listSize, nameList:sub(3), listTotal, listTotal/timePeriod, listTotal/listSize, listTotal/timePeriod/listSize) );
    elseif listSize == 1 then
        self:Print( string.format("%s got %0.0f extra damage from Blood Frenzy (%0.1f)", nameList:sub(3), listTotal, listTotal/timePeriod) )
    end
    if (numberDisplayed > 0 and playerDamage and playerDamage > 0) then
        self:Print( string.format("You got %0.0f extra damage from Blood Frenzy (%0.1f)", playerDamage, playerDamage/timePeriod) )
    end
end

function Slammer:PLAYER_TARGET_CHANGED()
    if UnitLevel("TARGET") == -1 then
        wantedList[UnitGUID("TARGET")] = true;
        --self:Print(UnitName("TARGET"), "added to wanted list") ;
    end    
end

function Slammer:PLAYER_REGEN_DISABLED()
    startTime = GetTime()
    spiritShockInterruptID = 1
end

function Slammer:ITEM_LOCK_CHANGED(arg1, arg2)
    if arg2 == nil and ( arg1 == 16 or arg1 == 17 or arg1 == 18) then
        self:NormalHit(true);        
    end
end

function Slammer:COMBAT_TEXT_UPDATE(larg1, larg2, larg3)
    if (larg1=="EXTRA_ATTACKS") and ( tonumber(larg2) >= 1 )then
        extraAttackTimeStamp = GetTime();
    end
end

function Slammer:NormalHit(userTimed)
    self:ScheduleEvent(function () self:AfterSwing("0.3") end, 0.3);
    self:ScheduleEvent(function () self:AfterSwing("0.8") end, 0.8);
    local id = "Slammer-Bar";
    local registered, time, elapsed, running = self:CandyBarStatus(id);
    self:DoBar(userTimed);
end

function Slammer:Parried()
    local id = "Slammer-Bar";
    local registered, time, elapsed, running = self:CandyBarStatus(id);
    if registered then
        local rage = UnitMana("player");
        local remainingTime = time - (GetTime() - hitTimeStamp);
        local twentyPercent = 0.2 * time
        if remainingTime > twentyPercent then -- that seem accurate
            self:SetCandyBarTimeLeft(id, math.max(remainingTime*0.6, twentyPercent) );  --the 40% is all over the place... seems we learn about parrying with much delay, which make this value unpredictable.
            expectedNextHitTime = expectedNextHitTime - math.min(remainingTime*0.4, remainingTime-twentyPercent); -- so think of it a a theoritical minimum in time reduction.
        end
        self:SetCandyBarText(id, string.format("Parried (%.02f) Rage: %i Offset: %.02f", time, rage, lastOffset));
    end
end

-- userTimed: if it is an attack timed by the user, i.e. true if slam 
function Slammer:DoBar(userTimed)
    local mainSpeed, offSpeed = UnitAttackSpeed("player");
    if offSpeed == nil then
        local id, duration, text, icon, color;
        id = "Slammer-Bar";
        icon = nil

        local rage = UnitMana("player");
        
        if not self:IsWindfuryProc() then
            if not userTimed then
                lastOffset = GetTime() - expectedNextHitTime;
            else
                -- wcf.Text:SetText(string.format("hit-slam delay %.02f", GetTime() - hitTimeStamp)); -- Moved to HitSlamDelay()
            end

            duration = mainSpeed
            hitTimeStamp = GetTime();
            expectedNextHitTime = hitTimeStamp + duration;

            text = string.format("(%.02f) Rage: %i Offset: %.02f", duration, rage, lastOffset);

            self:RegisterCandyBar(id, duration, text, icon, "navy", "olive", "navy", "red")
            self:RegisterCandyBarWithGroup(id, "Slammer")
            self:SetCandyBarWidth(id, self.db.profile.barwidth)
            self:SetCandyBarHeight(id, self.db.profile.barheight)
            if self.db.profile.reversebar then
                self:SetCandyBarReversed(id, true)
            end
            self:StartCandyBar(id, true)
        end
    end
end

function Slammer:HitSlamDelay()
    wcf.Text:SetText(string.format("hit-slam cast delay %.02f", GetTime() - hitTimeStamp));    
end

function Slammer:AfterSwing(arg)
    if self.db.profile.shownag then
        self:CheckForNag()
    end
    local mainSpeed, offSpeed = UnitAttackSpeed("player");
    local id = "Slammer-Bar";
    local registered, time, elapsed, running = self:CandyBarStatus(id);
    --self:Print("AS", arg, "reg time speed test:", registered, time, mainSpeed, mainSpeed ~= time)
    if registered and mainSpeed ~= time then
        --self:Print("Speed changed", arg)
        expectedNextHitTime = expectedNextHitTime - time + mainSpeed;
        local remainingTime = mainSpeed - (GetTime() - hitTimeStamp);
        self:SetCandyBarTime(id, mainSpeed)
        self:SetCandyBarTimeLeft(id, remainingTime)
    end
end

function Slammer:IsWindfuryProc()
    local timeSinceLastExtra = GetTime() - extraAttackTimeStamp
    if timeSinceLastExtra < 0.1 then
        return true
    end
    return false
end

local lastNag = GetTime()
function Slammer:CheckForNag()
    local NagCooldown = 3
    local _, playerClass = UnitClass("player")
    if playerClass == "WARRIOR" then
        if GetTime() > (NagCooldown + lastNag) then
            if UnitIsPlusMob("target") then
                if self.db.profile.tclaptime  ~= -1 then 
                    local tclap = self:TClapTime()
                    if tclap then
                        if tclap < self.db.profile.tclaptime then
                            self:Nag("Get TClap up.".." ("..math.floor(tclap*10)/10 .." left)")
                        end
                    else
                        self:Nag("Get TClap up.")
                    end
                end
                if self.db.profile.sundertime ~= -1 then 
                    local sunder = self:SunderFullTime() 
                    if sunder then
                        if sunder < self.db.profile.sundertime then
                            self:Nag("Refresh Sunders.".." ("..math.floor(sunder*10)/10 .." left)")
                        end
                    else
                        self:Nag("Help with Sunders."..self:SunderCount().."/5")
                    end
                end
                if self.db.profile.demotime   ~= -1 then 
                    local demo = self:DemoTime()
                    local demoWarning = ""
                    if self:HasCurseOfWeakness() then
                        demoWarning = " But mind Curse of Weakness"
                    end
                    if demo then
                        if demo < self.db.profile.demotime then
                            self:Nag("Get DemoShout up.".." ("..math.floor(demo*10)/10 .." left)"..demoWarning)
                        end
                    else
                        self:Nag("Get DemoShout up."..demoWarning)
                    end
                end
                if self.db.profile.bleedtime  ~= -1 and self:HasBloodFrenzy() then 
                    local bleed = self:BleedTime() 
                    if bleed then
                        if bleed < self.db.profile.bleedtime then
                            self:Nag("Get some bleed up.".." ("..math.floor(bleed*10)/10 .." left)")
                        end
                    else
                        self:Nag("Get some bleed up.")
                    end
                end
            end
            if self.db.profile.buffshouttime ~= -1 then
                local buffshout = self:BuffShoutTime()
                local shoutstring
                if self:IsBSFromOther() then 
                    shoutstring = "Get Commanding Shout up."
                else
                    shoutstring = "Get Battle Shout up."
                end
                    
                if buffshout then
                    if buffshout < self.db.profile.buffshouttime then
                        self:Nag(shoutstring.." ("..math.floor(buffshout*10)/10 .." left)")
                    end
                else
                    self:Nag(shoutstring)
                end
            end


            if self.db.profile.showdebuffcheck then
                self:VariousDebuffs();
            end
        end
    end
end

function Slammer:AddBFDamage(srcName,amount,dstGUID)
    local hasBloodFrenzy = self:HasBloodFrenzy()
    if BFActiveGUID[dstGUID] and hasBloodFrenzy then 
        if BFActiveGUID[dstGUID] > GetTime() then
            local gain = amount - amount / (1 + hasBloodFrenzy * 0.02);
            --self:Print("Adding " .. gain .. " damage from "..amount.." hit to BF damage.")
            BFDamage         = BFDamage +        gain;
            if not BFDamageSource[srcName] then
                BFDamageSource[srcName] = 0
            end
            --self:Print(string.format("Adding %0.1f damage from %d hit to BF damage to source %s (total was %0.1f)",gain, amount, srcName, BFDamageSource[srcName]))
            BFDamageSource[srcName] = BFDamageSource[srcName] + gain;                            
        else
            --self:Print("Adding " .. amount - amount * (1 + hasBloodFrenzy * 0.02) .. " damage from "..amount.." hit to lost damage.")
            BFDamageDowntime = BFDamageDowntime - amount + amount * (1 + hasBloodFrenzy * 0.02);
        end
    end
end

local maxBuffs = 32;
local maxDebuffs = 40;
local timeForAlreadyGone = 7;
local shoutDuration = 120; -- assuming no booming voice talent

function Slammer:SalvTankingCheck()
    if GetShapeshiftForm(true) == 2 then -- defensive stance
        for i=1,maxBuffs do
            index, untilCancelled = GetPlayerBuff(i, "HELPFUL")
            if (index > 0) then
                if string.find(GetPlayerBuffName(index),"Blessing of Salvation") then
                    self:Pour("Consider clicking off Salvation.",1.0,0.6,0.8);
                end
            end
        end
    end
end

function Slammer:VariousDebuffs()
    interestingDebuffs = "";
    somethingToSay = false;
    for i=1,maxDebuffs do
        name, _, _, count, _, _, timeLeft  = UnitDebuff("target", i)
        if name == "Curse of Recklessness" then
            interestingDebuffs = interestingDebuffs .. "CoR "
            somethingToSay = true;
        elseif name == "Demoralizing Roar" then
            interestingDebuffs = interestingDebuffs .. "DR "
            somethingToSay = true;
        elseif name == "Faerie Fire" or name == "Faerie Fire (Feral)" then
            interestingDebuffs = interestingDebuffs .. "FF "
            somethingToSay = true;
        elseif name == "Blood Frenzy" then
            interestingDebuffs = interestingDebuffs .. "BlFr "
            somethingToSay = true;
        elseif name == "Expose Weakness" then
            interestingDebuffs = interestingDebuffs .. "ExpW "
            somethingToSay = true;
        end

    end
    if somethingToSay then 
        lastNag = GetTime()    
        self:Pour(interestingDebuffs,1.0,0.9,0.8);
    end
    return
end

----
-- functions of that type return nil if the debuff is not up
function Slammer:TClapTime()
    for i=1,maxDebuffs do
        name, _, _, count, _, _, timeLeft  = UnitDebuff("target", i)
        if name == "Thunder Clap" then
            if timeLeft == nil then timeLeft = 30 end
            return timeLeft
        end
    end
    return nil
end

function Slammer:DemoTime()
    for i=1,maxDebuffs do
        name, _, _, count, _, _, timeLeft  = UnitDebuff("target", i)
        if name == "Demoralizing Shout" then
            if timeLeft == nil then timeLeft = 30 end
            return timeLeft
        end
    end
    return nil
end

function Slammer:HasCurseOfWeakness()
    for i=1,maxDebuffs do
        name, _, _, count, _, _, timeLeft  = UnitDebuff("target", i)
        if name == "Curse of Weakness" then
            return true
        end
    end
    return nil
end
    

function Slammer:SunderFullTime()
    for i=1,maxDebuffs do
        name, _, _, count, _, _, timeLeft  = UnitDebuff("target", i)
        if name == "Sunder Armor" then
            if count < 5 then
                return nil
            end
            if timeLeft == nil then timeLeft = 30 end
            return timeLeft
        end
    end
    return nil
end

function Slammer:BleedTime()
    local result = nil
    for i=1,maxDebuffs do
        name, _, _, count, _, _, timeLeft  = UnitDebuff("target", i)
        if name == "Deep Wound" or name == "Rend" then
            if timeLeft ~= nil then
                if result == nil then
                    result = timeLeft
                else
                    result = max(result,timeLeft)
                end
            end
        end
    end
    return result
end

function Slammer:SunderCount()
    for i=1,maxDebuffs do
        name, _, _, count, _, _, timeLeft  = UnitDebuff("target", i)
        if name == "Sunder Armor" then
            return count
        end
    end
    return 0
end

function Slammer:BuffShoutTime()
    if lastSelfShout == "BS" then 
        return self:BattleShoutTime()
    elseif lastSelfShout == "CS" then
        return self:CommandingShoutTime()
    end
    return nil
end

function Slammer:BattleShoutTime()
    for i=1,maxBuffs do
        index, untilCancelled = GetPlayerBuff(i, "HELPFUL")
        if (index > 0) then
            if GetPlayerBuffName(index) == "Battle Shout" then
                return GetPlayerBuffTimeLeft(index)
            end
        end
    end
    return nil
end

function Slammer:CommandingShoutTime()
    for i=1,maxBuffs do
        index, untilCancelled = GetPlayerBuff(i, "HELPFUL")
        if (index > 0) then
            if GetPlayerBuffName(index) == "Commanding Shout" then
                return GetPlayerBuffTimeLeft(index)
            end
        end
    end
    return nil
end

function Slammer:HasBloodFrenzy()
    local BloodFrenzyName,_,_,_,BloodFrenzyRank = GetTalentInfo(1,19)
    if BloodFrenzyName == "Blood Frenzy" and BloodFrenzyRank > 0 then 
        return BloodFrenzyRank
    else
        return nil
    end
end

function Slammer:HasDeepWounds()
    local DeepWoundsName = GetTalentInfo(1,9)
    if DeepWoundsName == "Deep Wounds" then 
        return true
    else
        return nil
    end
end

function Slammer:ShoutTalentedDuration()
    local BVName,_,_,_,BVRank = GetTalentInfo(2,1)
    if BVName == "Booming Voice" then
        return (BVRank*0.1+1)*120
    end
    return 120
end 

function Slammer:IsBSFromOther()
    if self:BattleShoutTime() and lastSelfShout ~= "BS" then
        return true
    else
        return false
    end
end

function Slammer:Nag(msg)
    lastNag = GetTime()    
    self:Pour(msg,0.6,1,1);
end



---- Non slammer-core function
-- function Slammer:INTERRUPT(envent, info)
--  if info.type == "interrupt" then
--      self:Print(string.format("Interrupted: %s from %s (%s)", info.skill, info.victim, info.source))
--  end
-- end
function Slammer:PreBossDied(dstName)
    HideUIPanel(FriendsFrame);
    HideUIPanel(CharacterFrame);
    self:ScheduleEvent(function () self:BossDied(dstName) end, 0.5);
end

function Slammer:BossDied(name)
    self:Print("Boss died:",name)
    PanelTemplates_SetTab(FriendsFrame, 5);
    ShowUIPanel(FriendsFrame);
    PanelTemplates_SetTab(CharacterFrame, 1);
    ShowUIPanel(CharacterFrame);
    if GetNumRaidMembers() == 0 then
        HideUIPanel(FriendsFrame);
        self:Print("Party composition:",UnitName("Party1"),UnitName("Party2"),UnitName("Party3"),UnitName("Party4"),UnitName("Party5"))
    end
    FriendsFrame_Update();
    Screenshot()
end
    

function Slammer:UpdateDrDamageMitigation(spellName, amount, critical, dstLevel)
--    self:CheckTable(DrDamage.options.args.Calculation.args.Mitigation)
    if DrDamage then 
        spellName = spellName == "Intercept Stun" and "Intercept" or spellName
        local originalMitigation = DrDamage.options.args.Calculation.args.Mitigation.get()
        local originalResilience = DrDamage.options.args.Calculation.args.Resilience.get()
        DrDamage.options.args.Calculation.args.Mitigation.set(0)
        DrDamage.options.args.Calculation.args.Resilience.set(0)
        local baseAmount = critical and select(2,DrDamage:MeleeCalc(spellName)).MinCrit or select(2,DrDamage:MeleeCalc(spellName)).Min
        local newMitigation = nil
        if baseAmount and baseAmount > 0 then 
            newMitigation = 100-amount/baseAmount*100
        
            -- 1 to 59
            -- %Reduction = (Armor / ([85 * Enemy_Level] + Armor + 400)) * 100
            -- 60 and up
            -- %Reduction = (Armor / ([467.5 * Enemy_Level] + Armor - 22167.5)) * 100
            if not dstLevel then 
                dstLevel = UnitLevel("PLAYER")
            elseif dstLevel == -1 then 
                dstLevel = UnitLevel("PLAYER") + 3
            end
            local denomFactor = dstLevel < 60 and 85*dstLevel + 400 or 467.5*dstLevel - 22167.5
            local apparentArmor = denomFactor/((100/newMitigation)-1)
        
            if math.abs(newMitigation - originalMitigation) > 1 then
                -- self:Print(string.format("DrDamage mitigation changed from %0.2f to %0.2f, apparent armor %d (level %d) with %d armor penetration.", originalMitigation, newMitigation, apparentArmor, dstLevel, GetArmorPenetration()))
                self:Print(string.format("Target armor: %d (%d apparent) %0.2f%% mitigation", apparentArmor + GetArmorPenetration(), apparentArmor, newMitigation))
            end
        end
        DrDamage.options.args.Calculation.args.Mitigation.set(newMitigation or originalMitigation)
        DrDamage.options.args.Calculation.args.Resilience.set(originalResilience)
    end
end

function Slammer:FlameBlastHappened()
    self:Print("Flame Blast happened");
    self:ScheduleEvent(function () self:Print("*** 8 seconds after Flame Blast, move up ***"); end, 8.0);
end
